"""Defining pandera schemas which will be used for dataframe validation."""

import pandera as pa

config_schema = pa.DataFrameSchema(
    {
        "SHORT": pa.Column(str),
        "LONG": pa.Column(str),
        "DATAPACKFILE": pa.Column(str),
        "CUSTOM_DESCRIPTION": pa.Column(str),
        "CUSTOM_GROUP": pa.Column(str),
    }
)

metadata_schema = pa.DataFrameSchema(
    {
        "sequential": pa.Column(str),
        "short": pa.Column(str),
        "long": pa.Column(str),
        "datapackfile": pa.Column(str),
        "profiletable": pa.Column(str),
        "columnheadingdescriptioninprofile": pa.Column(str),
    }
)



"""
censuswrangler package.

This package provides tools for extracting census data from the Australian Bureau of Statistics (ABS) datapack structure.
It includes utilities for managing datapack outputs using validated configuration templates.

Interface:
- create_config_template: Function to create a configuration template csv.
- Census: Class for managing, extracting and outputing census data from the datapack structure.
"""

from censuswrangler.config import create_config_template
from censuswrangler.census import Census

__all__ = ["create_config_template", "Census"]



"""
The census module provides the Census class, which encapsulates common configuration and datapack objects,
and provides methods to gather, filter, join, and output census data based on specified configurations.
"""

from typing import Dict, Optional
import copy
import datetime
import os

import pandas as pd

from censuswrangler._pack import pack
from censuswrangler._data import Data
from censuswrangler.config import Config


class Census:
    """
    The `Census` class is designed to facilitate the processing and manipulation of census data.
    It integrates configuration and datapack objects, enabling users to gather, filter, join,
    and output census data based on specified configurations. The class supports operations
    such as merging and pivoting dataframes, and provides methods to validate and output the
    processed data.

    Attributes:
        datapack_path (str): Path to the folder containing the census datapack.
        config_path (str): Path to the configuration file.
        geo_type (str): The spatial aggregation sub-folder to target (e.g., LGA, SA2).
        year (int): The census year used to identify columns in the datapack.
        col_type (str): The type of column output to use, either 'short' or 'long'. Defaults to 'short'.
        affix_type (str): Specifies whether to add a 'prefix', 'suffix', or 'none' to column names. Defaults to 'prefix'.
        config (Config): An instance of the `Config` class, representing the configuration file.
        pack (dict): A dictionary containing all the information needed to work on the datapack.
        data (Data): An instance of the `Data` class, built from the target geo and configuration.
        merged_df (Optional[pd.DataFrame]): A dataframe to store merged data after the `wrangle` method is called.
        pivoted_df (Optional[pd.DataFrame]): A dataframe to store pivoted data after the `wrangle` method is called.
    """

    def __init__(
        self,
        datapack_path: str,
        config_path: str,
        geo_type: str,
        year: int,
        col_type: str = "short",
        affix_type: str = "prefix",
    ):
        """
        Initializes the `Census` class with the specified datapack and configuration paths,
        geographic type, year, column type, and affix type. It also validates input parameters
        and prepares the necessary objects for data processing.

        Args:
            datapack_path (str): Path to the folder containing the census datapack.
            config_path (str): Path to the configuration file.
            geo_type (str): The spatial aggregation sub-folder to target (e.g., LGA, SA2).
            year (int): The census year used to identify columns in the datapack.
            col_type (str, optional): The type of column output to use, either 'short' or 'long'. Defaults to 'short'.
            affix_type (str, optional): Specifies whether to add a 'prefix', 'suffix', or 'none' to column names. Defaults to 'prefix'.

        Raises:
            AssertionError: If `col_type` is not one of the allowed values ('short', 'long').
            AssertionError: If `affix_type` is not one of the allowed values ('prefix', 'suffix', 'none').
        """
        # Where the census folder is
        self.datapack_path: str = datapack_path
        # Where the config file is saved
        self.config_path: str = config_path
        # What spatial aggregation sub-folder to target
        self.geo_type: str = geo_type
        # Helps find columns in the datapack, which have the census year as a suffix
        self.year: int = year
        # The type of column output to use. Can be 'short' or 'long'
        self.col_type: str = col_type
        # Affix a 'prefix', 'suffix' or 'none' of the csv's file code to each col, and put arg on file name
        self.affix_type: str = affix_type
        # Config object
        self.config: Config = Config(config_path)
        # All the info needed to work on the datapack
        self.pack: dict = pack(datapack_path)
        # Data object built from the target geo and config
        self.data: Data = Data(self.pack["data"]["path"], geo_type, self.config)
        # Allowed values for the output argument
        self._allowed_output_modes: Dict[Dict] = {
            "merge": {
                "requirement": "First run the Census.wrangle method with mode = 'merge'"
            },
            "pivot": {
                "requirement": "First run the Census.wrangle method with mode = 'pivot'"
            },
            "all": {
                "requirement": "First run the Census.wrangle method with mode = 'all'"
            },
        }

        # Basic parameter assertions
        allowed_col_types = ("short", "long")
        assert col_type in allowed_col_types, (
            f"col_type argument '{col_type} not in allowed types {allowed_col_types}"
        )

        allowed_affix_types = ("prefix", "suffix", "none")
        assert affix_type in allowed_affix_types, (
            f"affix_type argument '{affix_type} not in allowed types {allowed_affix_types}"
        )

        # Asserting that relevant config values exist in the metadata file of the pack
        def _assert_valid_metadata(config_list: list, metadata_col: str):
            """Common function for config metadata validation"""
            valid_entries = self.pack["metadata"]["files"]["metadata"]["columns"][
                metadata_col
            ]
            for entry in config_list:
                assert entry in valid_entries, (
                    f"'{entry}' in config file '{self.config.config_path_abs}' is not valid per datapack metadata in '{self.pack['metadata']['files']['metadata']['path']}', see sheet - 'Cell Descriptions Information'"
                )

        # Asserting each of the relevant config columns are valid
        _assert_valid_metadata(self.config.unique_short, "short")
        _assert_valid_metadata(self.config.unique_long, "long")
        _assert_valid_metadata(self.config.unique_datapackfile, "datapackfile")

        # Dataframes to store the merged and pivoted data (once wrangle is called)
        self.merged_df: Optional[pd.DataFrame] = None
        self.pivoted_df: Optional[pd.DataFrame] = None

    def _assert_mode_arg(self, mode):
        """Internal function to check the mode is in the allowed values"""
        allowed = self._allowed_output_modes.keys()
        assert mode in allowed, (
            f"mode argument '{mode}' not in allowed modes '{allowed}'"
        )

    def wrangle(self, mode):
        """
        Processes census data by gathering, filtering, and joining specified census files
        based on the configuration and datapack objects in the Census class.

        This method performs the following steps:

        1. Validates the `mode` argument to ensure it is one of the allowed values ('merge', 'pivot', 'all').
        2. Reads and filters data from census files based on the configuration.
        3. Renames and prepares columns according to the specified column type (`col_type`) and affix type (`affix_type`).
        4. Merges the prepared dataframes if the mode is 'merge' or 'all'.
        5. Creates pivoted dataframes grouped by specified column groups if the mode is 'pivot' or 'all'.

        Args:
            mode (str): The processing mode, which determines the type of operation to perform ('merge', 'pivot', 'all').

        Raises:
            AssertionError: If the `mode` argument is not one of the allowed values.
            ValueError: If invalid values are provided for `col_type` or `affix_type`.
        """
        self._assert_mode_arg(mode)

        # ===========
        # Prepare target dataframes
        # ===========

        # List to store column, name
        col_details = []

        # Looping through the per-file-code dictionaries, reading and filtering the resulting dataframes per the config
        for file_details in self.data.details:
            # Prepare the dataframe
            file_path = file_details["full_path"]
            unfiltered_df = pd.read_csv(file_path)
            file_details["unfiltered_df"] = unfiltered_df

            # Grab the current file code
            file_code = file_details["nameparts"]["file_code"]

            # Get the config, and select the rows that match the current file code
            # Save the df as a list of lists, where each list the values in the row
            df = self.config.df
            df = df[df["DATAPACKFILE"] == file_code]
            df = df.drop(columns=["DATAPACKFILE"])
            config_rows = df.values.tolist()

            # Store the column order
            short_index: int = df.columns.get_loc("SHORT")
            long_index: int = df.columns.get_loc("LONG")
            custom_description_index: int = df.columns.get_loc("CUSTOM_DESCRIPTION")
            custom_group_index: int = df.columns.get_loc("CUSTOM_GROUP")

            # Dictorary to store the old and new column names before renaming
            col_name_dict = {}

            # Looping over the list of config rows
            # Prepares a dictionary mainly used to create new column names depending on conditions
            for row in config_rows:
                # Getting variables from list
                col_short = row[short_index]  # SHORT
                col_name = row[long_index]  # LONG
                col_desc = row[custom_description_index]  # CUSTOM_DESCRIPTION
                col_group = row[custom_group_index]  # CUSTOM_GROUP

                # Setting the replacement column name conditionally depending on arguments
                if self.col_type == "short":
                    col_name = col_short
                elif self.col_type == "long":
                    col_name = col_name
                else:
                    raise ValueError(
                        "col_type must be either 'short or 'long' - incorrect value entered."
                    )

                # Adding a prefix or suffix depending on arguments
                if self.affix_type == "prefix":
                    col_name = file_details["nameparts"]["file_code"] + "_" + col_name
                elif self.affix_type == "suffix":
                    col_name = col_name + "_" + file_details["nameparts"]["file_code"]
                elif self.affix_type == "none":
                    # Leave var unchanged
                    col_name = col_name
                else:
                    raise ValueError(
                        "affix_type must be 'prefix', 'suffix' or 'none' - incorrect value entered."
                    )

                # Adding the old and new key combination to the outer dictionary
                col_name_dict[col_short] = col_name

                # Adding all column group dictionary to the associated list
                # Creating the dictionary
                col_detail = {
                    "old_col": col_short,
                    "new_col": col_name,
                    "group": col_group,
                    "col_desc": col_desc,
                }

                # Appending that to the list
                col_details.append(col_detail)

            # Getting a list with just the old col names (which are the keys)
            old_col_list = list(col_name_dict.keys())

            # Appending the target columns to the dictionary
            file_details["target_columns"] = col_name_dict

            # Establishing the name of the primary key column
            # This is basically the geocode with a suffix
            primary_key_col = f"{self.geo_type}_CODE_{self.year}"

            # Adding that to the list of old columns which is used to filter below
            old_col_list.insert(0, primary_key_col)

            # Renaming and filtering columns using the config data
            prepared_df = unfiltered_df.loc[:, old_col_list].rename(
                columns=col_name_dict
            )

            # Saving the prepared_df df to the file_details dict, which is in turn saved inplace to datapack.details
            file_details["prepared_df"] = prepared_df

        # ===========
        # Preparing outputs
        # ===========

        # ------------
        # Merge mode
        # ------------
        # Create an empty dataframe to store the merged data
        # Used in the pivot mode as well
        if mode == "merge" or mode == "pivot" or mode == "all":
            # Get all prepared dataframes in a list
            prepared_dfs = [detail["prepared_df"] for detail in self.data.details]

            # Loop through each dataframe in the list and merge with the 'merged_df'
            # Use the first df as the base and merge the rest on the primary key column
            for df in prepared_dfs:
                if self.merged_df is None:
                    self.merged_df = df
                else:
                    self.merged_df = pd.merge(
                        self.merged_df, df, on=primary_key_col, validate="one_to_one"
                    )

        # ------------
        # Pivot mode
        # ------------
        if mode == "pivot" or mode == "all":
            # Reworking the dictionary containing group and column information
            # Defining the new structure as a dict of lists like {'group': ['col1', 'col2', 'col3'],...}
            group_dict = {}

            for col_detail in col_details:
                group_key = col_detail["group"]
                new_col_value = col_detail["new_col"]
                if group_key not in group_dict:
                    group_dict[group_key] = []
                if new_col_value not in group_dict[group_key]:
                    group_dict[group_key].append(new_col_value)

            # Defining a list to contain output dataframes, which will be used to concat
            pivoted_dfs_list = []

            # Looping over the dictionary to subset, unpivot and create the new 'pivot' dataframes
            for (
                key_group,
                value_col_list,
            ) in (
                group_dict.copy().items()
            ):  # To avoid runtime errors to adding to a dict which being looped over
                # Creating a new list that includes the id column
                group_columns = value_col_list
                group_columns.append(primary_key_col)

                # Create a subset of the merged dataframe containing only columns from the group
                new_df = copy.deepcopy(self.merged_df[group_columns])

                # Creating a basic dictionary with the old (key) and new names (value)
                col_desc_dict = {}

                for ref_dict in col_details:
                    col_desc_dict[f"{ref_dict['new_col']}"] = ref_dict["col_desc"]

                # Using that dictionary to rename columns
                new_df = new_df.rename(columns=col_desc_dict)

                # Getting all columns that are not the primary key column for the pivoting function
                cols_to_unpivot = new_df.columns.difference([primary_key_col])

                # Unpivot dataframe
                new_df_unpivoted = new_df.melt(
                    id_vars=[primary_key_col],
                    value_vars=cols_to_unpivot,
                    var_name=key_group,
                    value_name=f"{key_group} Value",
                )

                # Appending those dataframes to the results list
                pivoted_dfs_list.append(new_df_unpivoted)

            # Concat-ing all unpivoted dfs
            pivot_concat_df = pd.concat(pivoted_dfs_list)

            # Assign it to the class attribute
            self.pivoted_df = pivot_concat_df

    def to_csv(self, mode: str, output_folder: str):
        """
        Outputs the processed census data to CSV files in the specified output folder.

        This method generates CSV files based on the specified mode ('merge', 'pivot', or 'all')
        and saves them to the provided output folder. The filenames include metadata such as
        geographic type, column type, affix type, and the current timestamp.

        Args:
            mode (str): The output mode, which determines the type of data to export ('merge', 'pivot', 'all').
            output_folder (str): The directory where the CSV files will be saved. Must be an existing directory.

        Raises:
            AssertionError: If the specified `output_folder` does not exist or is not a directory.
            ValueError: If the `mode` argument is not one of the allowed values ('merge', 'pivot', 'all').
        """

        self._assert_mode_arg(mode)

        # Check the folder directory
        assert os.path.isdir(output_folder), (
            f"The path '{output_folder}' is not a directory or does not exist."
        )

        # Common file name elements
        current_dt = datetime.datetime.now().strftime("%Y-%m-%d %H-%M")
        file_name_end = (
            "-"
            + self.geo_type
            + "_"
            + self.col_type
            + "_"
            + self.affix_type
            + "-"
            + current_dt
            + ".csv"
        )

        # Set names for the output types
        merge_name = "Census Data - Merge" + file_name_end
        pivot_name = "Census Data - Pivot" + file_name_end

        # Common csv output function
        def df_to_csv(df, name, index=False):
            df.to_csv(os.path.join(output_folder, name), index=index)

        # Conditionally Output the csv
        if mode == "merge":
            df_to_csv(self.merged_df, merge_name)
        elif mode == "pivot":
            df_to_csv(self.pivoted_df, pivot_name)
        elif mode == "all":
            df_to_csv(self.merged_df, merge_name)
            df_to_csv(self.pivoted_df, pivot_name)
        else:
            raise ValueError(
                f"'{mode}' is invalid. mode must be one of allowed values {self._allowed_output_modes.keys()}"
            )


if __name__ == "__main__":
    from icecream import ic

    census = Census(
        datapack_path=r"E:/Data/2021_GCP_all_for_AUS_short-header/",
        config_path=r"censuswrangler/config_template.csv",
        geo_type="LGA",
        year=2021,
    )

    census.wrangle("all")
    ic(census.merged_df.head(3))
    census.to_csv("all", r"F:/Github/censuswrangler/test_output")



"""Module for working with the config file."""

from io import StringIO
import os

from icecream import ic
import pandas as pd

from censuswrangler._schemas import config_schema


class Config:
    """Class for accessing the config file, providing useful attributes and checking the validity of the configuration."""

    def __init__(self, config_path: str):
        """Initialises the Config class object.

        Args:
            config_path (str): The file path of the config csv.
        """
        # Check if the config file exists then read it into a DataFrame
        self.config_path: str = config_path
        self.config_path_abs: str = os.path.abspath(config_path)
        assert os.path.exists(config_path), f"Config file not found at: {config_path}"
        self.df: pd.DataFrame = config_schema(pd.read_csv(config_path))

        # Cast all columns as strings
        self.df = self.df.astype(str)

        # Count rows
        self.row_count: int = len(self.df)
        self.unique_row_count: int = len(self.df.drop_duplicates())

        # Assert that there are no duplicate rows
        if self.row_count != self.unique_row_count:
            print("Duplicate rows found in config file, these will be ignored")

        # Get unique values for each column
        self.unique_datapackfile: list = self.df["DATAPACKFILE"].unique().tolist()
        self.unique_datapackfile_count: int = len(self.unique_datapackfile)

        self.unique_short: list = self.df["SHORT"].unique().tolist()
        self.unique_short_count: int = len(self.unique_short)

        self.unique_long: list = self.df["LONG"].unique().tolist()
        self.unique_long_count: int = len(self.unique_long)

        self.unique_custom_description: list = (
            self.df["CUSTOM_DESCRIPTION"].unique().tolist()
        )
        self.unique_custom_description_count: int = len(self.unique_custom_description)

        self.unique_custom_group: list = self.df["CUSTOM_GROUP"].unique().tolist()
        self.unique_custom_group_count: int = len(self.unique_custom_group)

    def summary(self):
        """Prints a summary of the config file"""
        ln = "-" * 50
        print(ln)
        print("Config file summary")
        print(ln)
        ic(self.config_path)
        ic(self.config_path_abs)
        ic(self.row_count)
        ic(self.unique_row_count)
        print(ln)
        ic(self.unique_datapackfile_count)
        ic(self.unique_short_count)
        ic(self.unique_long_count)
        ic(self.unique_custom_description_count)
        ic(self.unique_custom_group_count)
        print(ln)
        ic(self.unique_datapackfile)
        ic(self.unique_short)
        ic(self.unique_long)
        ic(self.unique_custom_description)
        ic(self.unique_custom_group)
        print(ln)


def create_config_template(
    output_folder: str = None, file_name: str = "censuswrangler_config"
) -> None:
    """Create a template config csv for use with the census

    Args:
        output_folder (str, optional): The folder path where the config file will be created. Defaults to the script location.
        file_name (str, optional): The name of the config file, excluding file type. Defaults to "censuswrangler_config".
    """
    # Determine is output_folder is blank
    output_folder_is_set = False
    if not (output_folder is None or output_folder == ""):
        output_folder_is_set = True

    # Allow none or blank but check folder if path is provided
    if output_folder_is_set:
        assert os.path.isdir(output_folder), (
            f"The provided folder_path argument '{output_folder}' is not a directory or does not exist."
        )

    # The whole config as a string, then converted into a csv
    csv_source = """SHORT,LONG,DATAPACKFILE,CUSTOM_DESCRIPTION,CUSTOM_GROUP
Tot_P_M,Total_Persons_Males,G01,Male,Gender
Tot_P_F,Total_Persons_Females,G01,Female,Gender
Age_0_4_yr_M,Age_groups_0_4_years_Males,G01,0 - 4 years,Age - Male
Age_0_4_yr_F,Age_groups_0_4_years_Females,G01,0 - 4 years,Age - Female
Age_0_4_yr_P,Age_groups_0_4_years_Persons,G01,0 - 4 years,Age - Person
Age_5_14_yr_M,Age_groups_5_14_years_Males,G01,5 - 14 years,Age - Male
Age_5_14_yr_F,Age_groups_5_14_years_Females,G01,5 - 14 years,Age - Female
Age_5_14_yr_P,Age_groups_5_14_years_Persons,G01,5 - 14 years,Age - Person
Age_15_19_yr_M,Age_groups_15_19_years_Males,G01,15 - 19 years,Age - Male
Age_15_19_yr_F,Age_groups_15_19_years_Females,G01,15 - 19 years,Age - Female
Age_15_19_yr_P,Age_groups_15_19_years_Persons,G01,15 - 19 years,Age - Person
Age_20_24_yr_M,Age_groups_20_24_years_Males,G01,20 - 24 years,Age - Male
Age_20_24_yr_F,Age_groups_20_24_years_Females,G01,20 - 24 years,Age - Female
Age_20_24_yr_P,Age_groups_20_24_years_Persons,G01,20 - 24 years,Age - Person
Age_25_34_yr_M,Age_groups_25_34_years_Males,G01,25 - 34 years,Age - Male
Age_25_34_yr_F,Age_groups_25_34_years_Females,G01,25 - 34 years,Age - Female
Age_25_34_yr_P,Age_groups_25_34_years_Persons,G01,25 - 34 years,Age - Person
Age_35_44_yr_M,Age_groups_35_44_years_Males,G01,35 - 44 years,Age - Male
Age_35_44_yr_F,Age_groups_35_44_years_Females,G01,35 - 44 years,Age - Female
Age_35_44_yr_P,Age_groups_35_44_years_Persons,G01,35 - 44 years,Age - Person
Age_45_54_yr_M,Age_groups_45_54_years_Males,G01,45 - 54 years,Age - Male
Age_45_54_yr_F,Age_groups_45_54_years_Females,G01,45 - 54 years,Age - Female
Age_45_54_yr_P,Age_groups_45_54_years_Persons,G01,45 - 54 years,Age - Person
Age_55_64_yr_M,Age_groups_55_64_years_Males,G01,55 - 64 years,Age - Male
Age_55_64_yr_F,Age_groups_55_64_years_Females,G01,55 - 64 years,Age - Female
Age_55_64_yr_P,Age_groups_55_64_years_Persons,G01,55 - 64 years,Age - Person
Age_65_74_yr_M,Age_groups_65_74_years_Males,G01,65 - 74 years,Age - Male
Age_65_74_yr_F,Age_groups_65_74_years_Females,G01,65 - 74 years,Age - Female
Age_65_74_yr_P,Age_groups_65_74_years_Persons,G01,65 - 74 years,Age - Person
Age_75_84_yr_M,Age_groups_75_84_years_Males,G01,75 - 84 years,Age - Male
Age_75_84_yr_F,Age_groups_75_84_years_Females,G01,75 - 84 years,Age - Female
Age_75_84_yr_P,Age_groups_75_84_years_Persons,G01,75 - 84 years,Age - Person
Age_85ov_M,Age_groups_85_years_and_over_Males,G01,> 85 years,Age - Male
Age_85ov_F,Age_groups_85_years_and_over_Females,G01,> 85 years,Age - Female
Age_85ov_P,Age_groups_85_years_and_over_Persons,G01,> 85 years,Age - Person
P_Tot_Marrd_reg_marrge,PERSONS_Total_Married_in_a_registered_marriage,G06,Married,Relationship Type
P_Tot_Married_de_facto,PERSONS_Total_Married_in_a_de_facto_marriage,G06,Couple,Relationship Type
P_Tot_Not_married,PERSONS_Total_Not_married,G06,No relationship,Relationship Type"""
    csv_df = pd.read_csv(StringIO(csv_source))

    # Prepare the output
    output_path = file_name + ".csv"
    if output_folder_is_set:
        output_path = os.path.join(output_folder, output_path)
    output_path_abs = os.path.abspath(output_path)

    # Check it isn't already there
    assert not os.path.exists(output_path), (
        f"File '{output_path_abs}' already exists, file creation aborted."
    )

    # Create the file and print the success message
    csv_df.to_csv(output_path, index=False)
    print(f"Successfully created censuswrangler config template: '{output_path_abs}'")


if __name__ == "__main__":
    from icecream import ic

    # # Example usage with the config template
    # config = Config("censuswrangler/config_template.csv")
    # config.summary()
    create_config_template("test_output")



"""Defines the pack() function, which builds a dictionary of information on the datapack."""

import os
import re

import pandas as pd

from censuswrangler._schemas import metadata_schema


def pack(
    folder_path: str,
    must_haves: list = ["Metadata", "Readme"],
    expectation_description: str = "Census Datapack folder",
    expected_folders: int = 3,
) -> dict[dict]:
    """Takes datapack path and builds a nested dictionary with info on each folder within.

    Args:
        folder_path (str): The path to the Datapack folder. This should be 3 Folders [ _data_, Metadata, Readme ]. The _data_ folder name can change depending on the downloaded datapack.
        must_haves (list[str], optional): Folders that must appear in the datapack. Defaults to ["Metadata", "Readme"].
        expectation_description (str, optional): An overall description of the expected Datapack structure. Defaults to "Census Datapack folder".
        expected_folders (_type_, optional): The total count of folders expected in the Datapack. Defaults to 3.

    Returns:
        dict[dict]: Nested dictionary of structured datapack information.
            data:
                name: str
                path: str
            metadata:
                name: str
                path: str
                files:
                    geographic_descriptions:
                        name: str
                        ext: str
                        path: str
                    sequential_template:
                        name: str
                        ext: str
                        path: str
                    metadata:
                        name: str
                        ext: str
                        path: str
                        df: DataFrame
                        columns:
                            sequential: list
                                List of unique values in the 'sequential' column.
                            short: list
                                List of unique values in the 'short' column.
                            long: list
                                List of unique values in the 'long' column.
                            datapackfile: list
                                List of unique values in the 'datapackfile' column.
                            profiletable: list
                                TList of unique values in the 'profiletable' column.
                            columnheadingdescriptioninprofile: list
                                List of unique values in the 'columnheadingdescriptioninprofile' column.
            readme:
                name: str
                path: str
    """

    def _check_dir(path: str) -> None:
        """Check that a string is a directory path.

        Args:
            path (str): The path which which should be a directory.
        """
        assert os.path.isdir(path), f"Error: '{path}' is not a directory."

    # Check the provided path
    _check_dir(folder_path)

    # Grab folder names and check them over
    folder_names = os.listdir(folder_path)
    folder_names_lower = [x.lower() for x in folder_names]
    folder_names_count = len(folder_names)
    # Run checks
    assert folder_names_count == expected_folders, (
        f"Expected '{folder_path}' to have {expected_folders} folders. Instead it has {folder_names_count}, please review. Description: {expectation_description}"
    )
    assert all(item.lower() in folder_names_lower for item in must_haves), (
        f"Required folders '{must_haves}' must all exist in folder '{folder_path}'"
    )

    # Gather info on each folder, build dict for each
    folder_info = {}
    for name in folder_names:
        # Get the full path
        path = os.path.join(folder_path, name)
        _check_dir(path)

        # Build the nested dict
        info = {}
        info["name"] = name
        info["path"] = path

        # Categorise the folder by type
        name_lower = name.lower()
        folder_type = ""
        if name_lower in ("metadata", "readme"):
            folder_type = name_lower
        else:
            folder_type = "data"

        # Add that to the outer dict by type
        folder_info[folder_type] = info

    # ----------------------------------------------
    # Build details on the files in the meta data folder
    # ----------------------------------------------

    # Initialise dict to contain info on the files of the metadata folder
    folder_info["metadata"]["files"] = {}

    # Loop through the metadata files and add them to that dict
    for root, dirs, files in os.walk(folder_info["metadata"]["path"]):
        for file_name_and_ext in files:
            # Split that out
            parts: tuple = os.path.splitext(file_name_and_ext)
            name: str = parts[0]
            ext: str = parts[1]
            path = os.path.join(root, name + ext)

            def assert_type(ext: str, expected_ext: str, file_desc: str):
                """Checks that that an extension is what was expected.

                Args:
                    ext (str): The file type to be checked.
                    expected_ext (str): The expected file type/ext, like 'csv', 'xlsx', or 'txt'.
                    file_desc (str): A description of the file being checked.
                """
                assert ext.lower() == expected_ext, (
                    f"Expected '{file_desc}' to be file type '{expected_ext}, instead got type '{ext}'"
                )

            # Identify the files, check their types
            file_desc = ""
            if bool(re.match(r"^Metadata_", name)):
                desc = "metadata"
                assert_type(ext, ".xlsx", desc)
                file_desc = desc
            elif bool(re.search(r"_Sequential_Template_", name)):
                desc = "sequential_template"
                assert_type(ext, ".xlsx", desc)
                file_desc = desc
            elif bool(re.search(r"_geog_desc_", name)):
                desc = "geographic_descriptions"
                assert_type(ext, ".xlsx", desc)
                file_desc = desc
            else:
                file_desc = name

            # Build a file details dict and put it into the pack dictionary
            file_info = {}
            file_info["name"] = name
            file_info["ext"] = ext
            file_info["path"] = path

            folder_info["metadata"]["files"][file_desc] = file_info

    # ----------------------------------------------
    # Extract the information in the cell descriptors sheet of the metadata xlsx
    # ----------------------------------------------

    # Read the Excel file without specifying the header
    metadata_file_path = folder_info["metadata"]["files"]["metadata"]["path"]
    sheet_names = pd.ExcelFile(metadata_file_path, engine="openpyxl").sheet_names

    # Find the sheet name that contains any combination of "cell" and "descriptors"
    def contains_cell_and_descriptors_in_sheet_name(sheet_name):
        """Check if the sheet name contains both "cell" and "descriptors" (case insensitive).

        Args:
            sheet_name (str): The name of the sheet to check.

        Returns:
            bool: True if the sheet name contains both "cell" and "descriptors", False otherwise.
        """
        return "cell" in sheet_name.lower() and "descriptors" in sheet_name.lower()

    sheet_name = next(
        (
            name
            for name in sheet_names
            if contains_cell_and_descriptors_in_sheet_name(name)
        ),
        None,
    )

    if sheet_name is None:
        raise ValueError(
            "No sheet name in the metadata xlsx contains 'cell' and 'descriptors'."
        )

    # Read the Excel file with the detected sheet name
    metadata_df = pd.read_excel(
        metadata_file_path,
        sheet_name=sheet_name,
        engine="openpyxl",
        header=None,
    )

    # Define a function to check if a row contains "Cell" and "Descriptors"
    def contains_cell_and_descriptors(row):
        """Check if a given row contains any cells with both the strings "Cell" and "Descriptors".

        Parameters:
        row (pd.Series): A pandas Series representing a row of data.

        Returns:
        bool: True if the row contains both "Cell" and "Descriptors", False otherwise.
        """
        contains_cell = row.astype(str).str.contains("Cell", case=False).any()
        contains_descriptors = (
            row.astype(str).str.contains("Descriptors", case=False).any()
        )
        return contains_cell and contains_descriptors

    # Find the row where "Cell" and "Descriptors" appear
    header_row_index = metadata_df.apply(contains_cell_and_descriptors, axis=1).idxmax()

    # Use the row after the found row as the header with values in lower case, and strip spaces
    metadata_df.columns = metadata_df.iloc[header_row_index + 1].str.lower()
    metadata_df.columns = metadata_df.columns.str.strip()

    # Drop the rows up to and including the header row
    metadata_df = metadata_df.drop(index=range(header_row_index + 2))

    # Reset the index of the DataFrame
    metadata_df = metadata_df.reset_index(drop=True)
    metadata_df.index = metadata_df.index + 1

    # Validate the metadata df
    validated_metadata_df = metadata_schema(metadata_df)

    # Start a dict with a key for each column containing a list of its unique values
    metadata_column_uniques = {
        col: metadata_df[col].unique().tolist() for col in metadata_df.columns
    }

    # Pack that up into its own dict, with the df
    metadata_info = {}
    metadata_info["df"] = validated_metadata_df
    metadata_info["columns"] = metadata_column_uniques

    # Add those to the dictionary
    folder_info["metadata"]["files"]["metadata"].update(metadata_info)

    return folder_info


if __name__ == "__main__":
    path = r"E:/Data/2021_GCP_all_for_AUS_short-header/"
    pack = pack(path)

    def print_dict_keys(d, indent=0):
        """Visualise the dictionary structure"""
        for key, value in d.items():
            print(" " * indent + str(key) + ":")
            if isinstance(value, dict):
                print_dict_keys(value, indent + 4)
            else:
                print(" " * (indent + 4) + "...")

    print_dict_keys(pack)
